﻿using System;
using System.Collections.Generic;
using Jypeli;
using Jypeli.Assets;
using Jypeli.Controls;
using Jypeli.Effects;
using Jypeli.Widgets;

public class Turtle : PhysicsGame
{
    // Kääntymiseen tarvittavat muuttujat
    Timer turnTimer = null;
    double turnSpeed = 250.0;
    Angle? turnTarget = null;

    // Liikkumiseen tarvittavat muuttujat
    Timer moveTimer = null;
    double moveSpeed = 250.0;
    Vector? moveTarget = null;

    GameObject currentObject = null;

    Boolean drawingOn = true;
    double drawingWidth = 5.0;
    Color lineColor = Color.Black;

    // Käskyt sisältävä jono
    Queue<Action> actionQueue = new Queue<Action>();

    PhysicsObject turtle;
    Image turtleKuva = LoadImage("turtle");

    /// <summary>
    /// Laitetaan kilpikonna liikkeelle.
    /// </summary>
    public override void Begin()
    {
        turtle = new PhysicsObject(99, 69);
        turtle.Image = turtleKuva;
        Add(turtle);

        IsMouseVisible = true;
        Level.Background.Color = Color.White;

        LisaaKaskyt();

        PhoneBackButton.Listen(ConfirmExit, "Lopeta peli");
        Keyboard.Listen(Key.Escape, ButtonState.Pressed, ConfirmExit, "Lopeta peli");

        Camera.ZoomFactor = 0.5;

        SeuraavaKasky();
    }

    /// <summary>
    /// Aliohjelma, jossa kilpikonnalle menevät käskyt lisätään.
    /// </summary>
    protected virtual void LisaaKaskyt()
    {
    }

    /// <summary>
    /// Liikutetaan kilpikonnaa pään osoittamaan suuntaan.
    /// </summary>
    /// <param name="pituus">Liikkeen pituus</param>
    private void MoveTo(double pituus)
    {
        if (moveTimer == null)
        {
            moveTimer = new Timer();
            moveTimer.Timeout += MoveToTarget;
            moveTimer.Interval = 0.001;
        }
        else if (moveTimer.Enabled)
            moveTimer.Stop();

        if (drawingOn)
        {
            currentObject = new GameObject(0, drawingWidth);
            currentObject.AbsolutePosition = turtle.AbsolutePosition;
            currentObject.AbsoluteAngle = turtle.AbsoluteAngle;
            currentObject.Color = lineColor;
            Add(currentObject);
        }
        moveTarget = turtle.AbsolutePosition + Vector.FromLengthAndAngle(pituus, turtle.AbsoluteAngle);
        moveTimer.Start();
    }

    /// <summary>
    /// Aliohjelma joka hoitaa kilpikonnan liikuttamisen.
    /// </summary>
    private void MoveToTarget()
    {
        if (!moveTarget.HasValue)
        {
            moveTimer.Stop();
            return;
        }

        Vector d = moveTarget.Value - turtle.AbsolutePosition;
        double vt = moveSpeed * moveTimer.Interval;

        if (d.Magnitude < vt)
        {
            moveTimer.Stop();
            turtle.AbsolutePosition = moveTarget.Value;
            moveTarget = null;
            SeuraavaKasky();
            if (drawingOn)
            {
                currentObject.Size += new Vector(d.Magnitude, 0);
                currentObject.AbsolutePosition += Vector.FromLengthAndAngle(d.Magnitude / 2, d.Angle);
            }
        }
        else
        {
            if (drawingOn)
            {
                currentObject.Size += new Vector(vt, 0);
                currentObject.AbsolutePosition += Vector.FromLengthAndAngle(vt / 2, d.Angle);
            }
            turtle.AbsolutePosition += Vector.FromLengthAndAngle(vt, d.Angle);
        }
    }

    /// <summary>
    /// Käännetään kilpikonnaa. Kilpikonna kääntyy vastapäivään, eli jos haluat
    /// kääntää kilpikonnaa myötäpäivään esim. 90 astetta, on se silloin 270
    /// astetta. (360-90=270)
    /// </summary>
    /// <param name="asteet">Kulma asteina</param>
    private void TurnTo(double asteet)
    {
        if (turnTimer == null)
        {
            turnTimer = new Timer();
            turnTimer.Timeout += TurnToTarget;
            turnTimer.Interval = 0.001;
        }
        else if (turnTimer.Enabled)
            turnTimer.Stop();

        turnTarget = turtle.Angle + Angle.FromDegrees(asteet);
        turnTimer.Start();
    }

    /// <summary>
    /// Aliohjelma, joka hoitaa kilpikonnan kääntämisen.
    /// </summary>
    private void TurnToTarget()
    {
        if (!turnTarget.HasValue)
        {
            turnTimer.Stop();
            return;
        }

        Angle angleLeft = turnTarget.Value - turtle.Angle;
        Angle angleSingleInterval = Angle.FromDegrees(turnSpeed) * turnTimer.Interval;

        if (angleLeft.GetPositiveDegrees() < angleSingleInterval.GetPositiveDegrees())
        {
            turnTimer.Stop();
            turtle.Angle = turnTarget.Value;
            turnTarget = null;
            SeuraavaKasky();
        }
        else
        {
            turtle.AbsoluteAngle += angleSingleInterval;
        }
    }

    /// <summary>
    /// Laittaa piirtämisen päälle.
    /// </summary>
    private void DrawingOn()
    {
        drawingOn = true;
        SeuraavaKasky();
    }

    /// <summary>
    /// Ottaa piirtämisen pois päältä.
    /// </summary>
    private void DrawingOff()
    {
        drawingOn = false;
        SeuraavaKasky();
    }
    
    /// <summary>
    /// Asettaa piirtojäljen paksuuden.
    /// </summary>
    /// <param name="paksuus">Kuinka paksua jälkeä piirretään</param>
    private void DrawingWidth(double paksuus)
    {
        drawingWidth = paksuus;
        SeuraavaKasky();
    }

    /// <summary>
    /// Asettaa piirtojäljen värin.
    /// </summary>
    /// <param name="color">Minkä väristä jälkeä piirretään</param>
    private void DrawingColor(Color color)
    {
        lineColor = color;
        SeuraavaKasky();
    }

    /// <summary>
    /// Asettaa piirtonopeuden
    /// </summary>
    /// <param name="speed">Piirtonopeus</param>
    private void DrawingSpeed(double speed)
    {
        moveSpeed = turnSpeed = speed;
        SeuraavaKasky();
    }

    /// <summary>
    /// Siirtää kilpikonnan suoraan annettuun paikkaan.
    /// </summary>
    /// <param name="target">Määränpää</param>
    private void MoveInstantly(Vector target)
    {
        if (drawingOn)
        {
            currentObject = new GameObject(Vector.Distance(turtle.Position, target), drawingWidth);
            currentObject.AbsolutePosition = (turtle.Position +  target) / 2;
            currentObject.AbsoluteAngle = (turtle.Position - target).Angle;
            currentObject.Color = lineColor;
            Add(currentObject);
        }

        turtle.AbsolutePosition = target;
        SeuraavaKasky();
    }

    /// <summary>
    /// Aliohjelma seuraavan käskyn suorittamista varten. Kutsutaan aina kun
    /// yksi käsky on saatu suoritettua loppuun.
    /// </summary>
    private void SeuraavaKasky()
    {
        if (actionQueue.Count <= 0)
            return;
        Action seuraava = actionQueue.Dequeue();
        seuraava();
    }

    /// <summary>
    /// Liikuttaa kilpikonnaa suoraan.
    /// </summary>
    /// <param name="pituus">Matka, jonka kilpikonna kulkee</param>
    public void Liiku(double pituus)
    {
        actionQueue.Enqueue(delegate { MoveTo(pituus); });
    }


    /// <summary>
    /// Liikuttaa kilpikonnan annettuun paikkaan.
    /// </summary>
    /// <param name="x">Paikan x-koordinaatti</param>
    /// <param name="y">Paikan y-koordinaatti</param>
    public void LiikuPaikkaan(double x, double y)
    {
        Vector v = new Vector(x, y);
        actionQueue.Enqueue(delegate {
            TurnTo(((v - turtle.Position).Angle.Degrees - turtle.Angle.Degrees) % 360 ); });
        actionQueue.Enqueue(delegate {
            MoveTo(Vector.Distance(v, turtle.Position)); });
    }


    /// <summary>
    /// Siirtää kilpikonnan välittömästi annettuun paikkaan. Ei käännä kilpikonnaa.
    /// </summary>
    /// <param name="x">Paikan x-koordinaatti</param>
    /// <param name="y">Paikan y-koordinaatti</param>
    public void Pomppaa(double x, double y)
    {
        actionQueue.Enqueue(delegate { MoveInstantly(new Vector(x, y)); });
    }

    /// <summary>
    /// Kääntää kilpikonnaa. Kilpikonna kääntyy vastapäivään, eli jos haluat
    /// kääntää kilpikonnaa myötäpäivään esim. 90 astetta, on se silloin 270
    /// astetta. (360-90=270)
    /// </summary>
    /// <param name="kulma">Kulma asteina</param>
    public void Kaanny(double kulma)
    {
        actionQueue.Enqueue(delegate { TurnTo(kulma); });
    }

    /// <summary>
    /// Laittaa kilpikonnan piirtämään viivaa peräänsä.
    /// </summary>
    public void PiirtoPaalle()
    {
        actionQueue.Enqueue(DrawingOn);
    }

    /// <summary>
    /// Ottaa kilpikonnan piirtämän viivan pois päältä.
    /// </summary>
    public void PiirtoPois()
    {
        actionQueue.Enqueue(DrawingOff);
    }

    /// <summary>
    /// Piirtää minkä tahansa Jypelin valmiista muodoista täytettynä piirtovärillä.
    /// </summary>
    /// <param name="muoto">Piirrettävä muoto</param>
    /// <param name="leveys">Muodon leveys</param>
    /// <param name="korkeus">Muodon korkeus</param>
    public void PiirraMuoto(Shape muoto, double leveys, double korkeus)
    {
        actionQueue.Enqueue(delegate {
            if (drawingOn)
            {
                GameObject p = new GameObject(leveys, korkeus, muoto);
                p.Position = turtle.Position;
                p.Angle = turtle.Angle; p.Color = lineColor; Add(p);
            }
            SeuraavaKasky();
        });
    }

    /// <summary>
    /// Asettaa kilpikonnan piirtämän viivan paksuuden.
    /// </summary>
    /// <param name="paksuus">Kunka paksua viivaa piirretään</param>
    public void AsetaViivanPaksuus(double paksuus)
    {
        actionQueue.Enqueue(delegate { DrawingWidth(paksuus); });
    }

    /// <summary>
    /// Asettaa kilpikonnan piirtämän viivan värin.
    /// </summary>
    /// <param name="vari">Piirrettävän viivan väri</param>
    public void AsetaVari(Color vari)
    {
        actionQueue.Enqueue(delegate { DrawingColor(vari); });
    }

    /// <summary>
    /// Asettaa kentän taustan värin.
    /// </summary>
    /// <param name="vari">Taustan väri</param>
    public void AsetaTaustavari(Color vari)
    {
        actionQueue.Enqueue(delegate { Level.Background.Color = vari; SeuraavaKasky(); });
    }

    /// <summary>
    /// Asettaa kilpikonnan piirtonopeuden
    /// </summary>
    /// <param name="nopeus">Piirtonopeus</param>
    public void AsetaNopeus(double nopeus)
    {
        actionQueue.Enqueue(delegate { DrawingSpeed(nopeus); });
    }

    /// <summary>
    /// Piilottaa kilpikonnan.
    /// </summary>
    public void Piilota()
    {
        actionQueue.Enqueue(delegate { turtle.IsVisible = false; SeuraavaKasky(); });
    }

    /// <summary>
    /// Palauttaa kilpikonnan näkyviin.
    /// </summary>
    public void Nayta()
    {
        actionQueue.Enqueue(delegate { turtle.IsVisible = true; SeuraavaKasky(); });
    }
}
